package org.sat4j.reader;

import java.io.IOException;
import java.io.LineNumberReader;
import java.util.Scanner;

import org.sat4j.reader.csp.Nogoods;
import org.sat4j.reader.csp.Relation;
import org.sat4j.reader.csp.SupportsDirectEncoding;
import org.sat4j.reader.csp.Var;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.IProblem;
import org.sat4j.specs.ISolver;

/**
 * This class is a CSP to SAT translator that is able to read
 * a CSP problem using the First CSP solver competition input 
 * format and that translates it into clausal and cardinality 
 * (equality) constraints.
 * 
 * That code has not been tested very thoroughtly yet and was 
 * written very quickly to meet the competition deadline :=))
 * There is plenty of room for improvement.
 * 
 * @author leberre
 * 
 */
public class CSPReader extends Reader {

    private final ISolver solver;

    private Var[] vars;

    protected Relation[] relations;

    public CSPReader(ISolver solver) {
        this.solver = solver;
    }
    
	@Override
	public final IProblem parseInstance(final java.io.Reader in) 
		throws ParseFormatException, ContradictionException, IOException {		
		return parseInstance(new LineNumberReader(in));
		
	}

    private IProblem parseInstance(LineNumberReader in) throws ParseFormatException,
            ContradictionException {
        solver.reset();
        try {
            readProblem(in);
            return solver;
        } catch (NumberFormatException e) {
            throw new ParseFormatException("integer value expected on line "
                    + in.getLineNumber(), e);
        }
    }

    @Override
    public String decode(int[] model) {
        StringBuilder stb = new StringBuilder();
        for (int i = 0; i < vars.length; i++) {
            stb.append(vars[i].findValue(model));
            stb.append(" ");
        }
        return stb.toString();
    }

    private void readProblem(LineNumberReader in) throws ContradictionException {
        Scanner input = new Scanner(in);
        // discard problem name
        System.out.println("c reading problem named "+input.nextLine());
        // read number of domain
        System.out.print("c reading domains");
        int nbdomain = input.nextInt();
        int [] [] domains = new int[nbdomain][];
        for (int i = 0; i < nbdomain; i++) {
            // read domain number
            int dnum = input.nextInt();
            assert dnum == i;

            // read domain
            domains[dnum] = readArrayOfInt(input);
        }
        System.out.println(" done.");
        System.out.print("c reading variables");
        int nbvar = input.nextInt();
        vars = new Var[nbvar];
        int nbvarstocreate=0;
        for (int i = 0; i < nbvar; i++) {            
            // read var number
            System.out.print("\rc reading variables "+i+"/"+nbvar);
            int varnum = input.nextInt();
            assert varnum==i;
            // read var domain
            int vardom = input.nextInt();
            vars[varnum] = new Var(domains[vardom], nbvarstocreate);
            nbvarstocreate+=domains[vardom].length;
        }    
        System.out.println("\rc reading variables ("+nbvar+") done.");
        solver.newVar(nbvarstocreate);
        for (int i = 0; i < nbvar; i++) {
            vars[i].toClause(solver);
        }
        System.out.print("c reading relations");
        // relation definition
        // read number of relations
        int nbrel = input.nextInt();
        relations = new Relation[nbrel];
        for (int i = 0; i < nbrel; i++) {
            System.out.print("\rc reading relations "+i+"/"+nbrel);
            // read relation number
            int relnum = input.nextInt();
            assert relnum == i;

            boolean forbidden = input.nextInt() == 1 ? false : true;
            int[] rdomains = readArrayOfInt(input);

            int nbtuples = input.nextInt();
            if (forbidden) {
                relations[relnum] = new Nogoods(rdomains, nbtuples);
            } else {
                manageAllowedTuples(relnum, rdomains, nbtuples);
            }
            // allowed/forbidden tuples
            for (int j = 0; j < nbtuples; j++) {
                int[] tuple = readArrayOfInt(input, relations[relnum].arity());
                // do something with tuple
                relations[relnum].addTuple(j, tuple);
            }
        }
        System.out.println("\rc reading relations ("+nbrel+") done.");
        System.out.print("c reading constraints");
        // constraint definition
        int nbconstr = input.nextInt();
        for (int i = 0; i < nbconstr; i++) {
            System.out.print("\rc reading constraints "+i+"/"+nbconstr);
            int[] variables = readArrayOfInt(input);
            int relnum = input.nextInt();
            // manage constraint
            relations[relnum].toClause(solver, intToVar(variables));
        }
        System.out.println("\rc reading constraints ("+nbconstr+") done.");
    }

    protected void manageAllowedTuples(int relnum, int[] domains, int nbtuples) {
        relations[relnum] = new SupportsDirectEncoding(domains, nbtuples);
    }

    private Var[] intToVar(int[] variables) {
        Var[] nvars = new Var[variables.length];
        for (int i = 0; i < variables.length; i++)
            nvars[i] = vars[variables[i]];
        return nvars;
    }

    private int[] readArrayOfInt(Scanner input) {
        int size = input.nextInt();
        return readArrayOfInt(input, size);
    }

    private int[] readArrayOfInt(Scanner input, int size) {
        int[] tab = new int[size];
        for (int i = 0; i < size; i++)
            tab[i] = input.nextInt();
        return tab;
    }
}
