package org.sat4j.tools;

import org.sat4j.core.VecInt;
import org.sat4j.specs.ContradictionException;
import org.sat4j.specs.IConstr;
import org.sat4j.specs.ISolver;
import org.sat4j.specs.IVecInt;
import org.sat4j.specs.TimeoutException;

/**
 * This solver decorator allows to detect whether or not the set of constraints 
 * available in the solver has only one solution or not.
 * 
 * NOTE THAT THIS DECORATOR CANNOT BE USED WITH SOLVERS USING SPECIFIC 
 * DATA STRUCTURES FOR BINARY OR TERNARY CLAUSES!
 * 
 *  <code>
   SingleSolutionDetector problem = 
     new SingleSolutionDetector(SolverFactory.newMiniSAT());
   // feed problem/solver as usual

   if (problem.isSatisfiable()) {
	if (problem.hasASingleSolution()) {
	      // great, the instance has a unique solution
              int [] uniquesolution = problem.getModel();
        } else {
              // too bad, got more than one
        }
   }
 *  </code>
 * @author leberre
 *
 */
public class SingleSolutionDetector extends SolverDecorator {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;

	public SingleSolutionDetector(ISolver solver) {
		super(solver);
	}

	/**
	 * Please use that method only after a positive answer from 
	 * isSatisfiable() (else a runtime exception will be launched).
	 * 
	 * @return true iff there is only one way to satisfy all the constraints 
	 * in the solver.
	 * @throws TimeoutException
	 */
	public boolean hasASingleSolution() throws TimeoutException {
		return hasASingleSolution(new VecInt());
	}

	/**
	 * Please use that method only after a positive answer from 
	 * isSatisfiable(assumptions) (else a runtime exception will be launched).
	 * 
	 * @param assumptions a set of literals (dimacs numbering) that must be 
	 * satisfied.
	 * @return true iff there is only one way to satisfy all the constraints 
	 * in the solver using the provided set of assumptions.
	 * @throws TimeoutException
	 */
	public boolean hasASingleSolution(IVecInt assumptions)
			throws TimeoutException {
		int [] firstmodel = model();
		assert firstmodel != null;
		IVecInt clause = new VecInt(firstmodel.length);
		for (int q : firstmodel) {
			clause.push(-q);
		}
		boolean result = false;
		try {
			IConstr added = addClause(clause);
			result = !isSatisfiable(assumptions);
			removeConstr(added);
		} catch (ContradictionException e) {
			result = true;
		}
		return result;
	}
}
